#include <u.h>
#include <libc.h>
#include <String.h>
#include <auth.h>
#include <authsrv.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include "httpd.h"
#include "httpsrv.h"

typedef struct Buf	Buf;
typedef struct Hdr	Hdr;
typedef struct Tab	Tab;

struct Buf
{
	char	*s;
	long	n;
};

struct Tab
{
	char 		*s;
	int		i;
	int		m;
	Tab		*t;
};

struct Hdr
{
	HSPairs	*ct;
	HSPairs	*cd;
	HSPairs	*a;
};

enum
{
	Xctl = 1,
	Xlog,
	/* request */
	Xreqbody,
	Xreqinfo,
	Xreqmethod,
	Xreqname,
	Xreqpass,
	Xreqpath,
	Xreqquery,
	Xreqrmtaddr,
	Xrequser,
	/* response */
	Xresbody,
	Xresctype,
	Xrescookie,
	Xresnph,
	Xresreply,
	NBUFS,
};

Tab authtab[]= {
	"user",				Xrequser,			0444,		nil,
	"pass",				Xreqpass,			0444,		nil,
	nil,					0,				0,			nil
};

Tab reqtab[]= {
	"auth",				0,				0555|DMDIR,	authtab,
	"body",				Xreqbody,			0444,		nil,
	"info",				Xreqinfo,			0444,		nil,
	"method",				Xreqmethod,		0444,		nil,
	"multipart",			0,				0555|DMDIR,	nil,
	"name",				Xreqname,		0444,		nil,
	"path",				Xreqpath,			0444,		nil,
	"query",				Xreqquery,		0444,		nil,
	"remote-addr",			Xreqrmtaddr,		0444,		nil,
	nil,					0,				0,			nil,
};

Tab restab[] = {
	"body",				Xresbody,			0666,		nil,
	"content-type",		Xresctype,		0666,		nil,
	"cookie",				Xrescookie,		0666,		nil,
	"nph",				Xresnph,			0666,		nil,
	"reply",				Xresreply,			0666,		nil,
	nil,					0,				0,			nil
};

Tab roottab[] = {
	"ctl",					Xctl,				0666,		nil,
	"log",				Xlog,				0222,		nil,
	"request",				0,				0555|DMDIR,	reqtab,
	"response",			0,				0777|DMDIR,	restab,
	nil,					0,				0,			nil,
};

char cgibinpath[] = "/usr/sns/bin/cgis";
char	uid[] = "cgifs";
char	cgimtpt[] = "/mnt/cgi";
char Eexec[] = "magic/cgifs: exec failed";
char Einternal[] = "500 Internal Error";
char	Tformenc[] = "application/x-www-form-urlencoded";
char	Tmultipart[] = "multipart/form-data";

HConnect	*cgihc;
int	nmultipart;
Buf	*buffer;
int	nbuffer;
Hdr	reqhdr;

void	fsread(Req *);
void	fswrite(Req *);

Srv fs = {
.read =	fsread,
.write =	fswrite,
};

void
error(char *fmt)
{
	Hio *h;
	char buf[1024];
	int n;

	h = &cgihc->hout;
	n = snprint(buf, sizeof(buf), 
		"<head><title>%s</title></head>\n<body><h1>%s</h1><p>%s</p></body>\n",
		Einternal, Einternal, fmt);
	hprint(h, "%s %s\r\n", hversion, Einternal);
	hprint(h, "Server: Plan9\r\n");
	hprint(h, "Content-Length: %d\r\n", n);
	if(cgihc->head.closeit)
		hprint(h, "Connection: close\r\n");
	else if(!http11(cgihc))
		hprint(h, "Connection: Keep-Alive\r\n");
	hprint(h, "Content-type: text/html\r\n");
	hprint(h, "\r\n");
	hwrite(h , buf, n);
	hflush(h);

	postnote(PNGROUP, getpid(), "kill");
	_exits(fmt);
}

char *
getval(HSPairs *p, char *s)
{
	for(; p; p=p->next)
		if(strncmp(p->s, s, strlen(s)) == 0)
			return p->t;
	return nil;
};

Buf *
bufget(int i)
{
	assert(i < nbuffer);
	return &buffer[i];
}

void
bufset(int i, char *s, long n)
{
	Buf *b;

	b = &buffer[i];
	b->s = s;
	b->n = n;
}

int
bufnew(char *s, long n)
{
	buffer = erealloc9p(buffer, ++nbuffer*sizeof(Buf));
	bufset(nbuffer-1, s, n);
	return nbuffer-1;
}

void
parsepath(void)
{
	char *s, *p, *name, *path, *info;

	s = hstrdup(cgihc, cgihc->req.uri);	/* actually just the path */
	if(s[0] != '/')
		error("parse path");

	name = s+1;
	p = strchr(name+1, '/');
	if(p == nil)
		info = "";
	else{
		info = hstrdup(cgihc, p);
		*p = '\0';
	}
	/*
	 * since the script name contains no slashes except the beginning one,
	 * it's safe.  at worst it's "/.." or "/.", which will fail when we exec.
	 */
	path = smprint("%s%s", cgibinpath, s);
	if(path == nil)
		error("out of memory");
	bufset(Xreqinfo, info, strlen(info));
	bufset(Xreqname, name, strlen(name));
	bufset(Xreqpath, path, strlen(path));
}

int
isspace(char c)
{
	return c==' ' || c== '\t' || c=='\r' || c=='\n';
}

HSPairs *
parsevalues(char *s)
{
	HSPairs *pair;
	char *p, *q;

	pair = hmkspairs(cgihc, hstrdup(cgihc, "type"), s, nil);
	p = s;
	while((p=strchr(p, ';')) != nil){
		*p++ = '\0';
		while(isspace(*p))
			p++;
		q = strchr(p, '=');
		*q++ = '\0';
		pair = hmkspairs(cgihc, p, q, pair);
		p = q;
	}
	return hrevspairs(pair);
}

void
parseheader(Hdr *h, char *os, int len)
{
	char *s, *f, *v, *p, *e;

	memset(h, 0, sizeof(Hdr));
	s = emalloc9p(len+1);
	memmove(s, os, len);
	s[len] = '\0';
	for(p=s; p && *p; ){
		f = p;
		while((p = strchr(p, '\n')) != nil){
			if(p > s && p[-1] == '\r')
				p[-1] = ' ';
			if(p[1] != ' ' && p[1] != '\t'){
				*p++ = '\0';
				break;
			}
			*p++ = ' ';
		}
		v = strchr(f, ':');
		if(v == nil)
			continue;
		*v++ = '\0';
		while(*v == ' ' || *v == '\t')
			v++;
		e = v+strlen(v);
		while(e > v && (e[-1]==' ' || e[-1]=='\t'))
			*--e = '\0';
		f = hlower(f);
		if(strncmp(f, "content-type", 12)  == 0)
			h->ct = parsevalues(v);
		else if(strncmp(f, "content-disposition", 19)  == 0)
			h->cd = parsevalues(v);
		else if(strncmp(f, "cookie", 6) != 0) /* ignore cookies, they are in cgih->head.cookie */
			h->a = hmkspairs(cgihc, f, v, h->a);
	}
	h->a = hrevspairs(h->a);
}

/*
 * we allocate memory for Xreqbody one,
 * so we don't worry about the offsets
*/
static Hio *hpost;

int
getpost(void)
{
	static int first = 1;
	char *s;
	long n;
	Buf *b;

	b = bufget(Xreqbody);
	if(first){
		hpost = hbodypush(&cgihc->hin, cgihc->head.contlen, cgihc->head.transenc);
		if(hpost == nil)
			return 0;
		n = atoi(getval(reqhdr.a, "content-length"));
		if(n < 0)
			return 0;
		b->s = emalloc9p(n+1);
		first = 0;
	}
	s = hreadbuf(hpost, hpost->pos);
	if(s == nil)
		return 0;
	n = hbuflen(hpost, s);
	memmove(b->s+b->n, s, n);
	b->n  += n;
	b->s[b->n] = '\0';
	hpost->pos = hpost->stop;
	return 1;
}

char *
findstr(char *start, char *end, char *s)
{
	char *p;
	int offset, n;

	offset = 0;
	n = strlen(s);
	for(p=start; p<end; p++){
		if(*p==s[offset]){
			offset++;
			if(offset == n)
				return p-n;
		}else if(offset > 0)
			offset = 0;
	}
	return nil;
}

void
parsemultipart(Hdr *h, int n, char **sp, long *np)
{
	/* rfc1341: boundaries can't be longer than 70 */
	static char boundary[128], *p;
	Buf *rb;
	char *q, *bs;

	bs = nil;
	rb = bufget(Xreqbody);
	if(n == 0){
		if((q=getval(reqhdr.ct, "boundary")) == nil)
			error("no boundary");
		sprint(boundary, "--%s", q);
		if(rb->s==nil && !getpost())
			error("no post body");
		p = rb->s;
	}

    l0:
	q = findstr(p, rb->s+rb->n, boundary);
	/* no boundary, done  */
	if(q == nil){
		if(getpost())
			goto l0;
		return;
	}
	if(bs){
		*sp = bs;
		*np = q-bs;
		return;
	}
	q += strlen(boundary);
	/* last part? */
	if(strncmp(q, "--", 2) == 0)
		return;
    l1:
	p = strchr(q, '\n');
	if(p == nil || (p+1)==nil){
		if(getpost())
			goto l1;
		return;
	}
    l2:
	q = strstr(p, "\r\n\r\n");
	if(q == nil){
		if(getpost())
			goto l2;
		return;
	}
	q += 4; /* eat "\r\n\r\n" */
	parseheader(h, p, q-p);
	bs = q;
	p = q;
	goto l0;
}

void
mkdirtab(File *d, Tab *t)
{
	File *f;

	for(; t->s; t++){
		f = createfile(d, t->s, uid, t->m, (void *)(t->i));
		if(t->t) mkdirtab(f, t->t);
		closefile(f);
	}
}

void
mkpair(File *d, HSPairs *p, char *s)
{
	if(s) d = createfile(d, s, uid, 0555|DMDIR, (void *)0);
	if(d == nil) error("can't create dir");
	for(; p; p=p->next)
		closefile(createfile(d, p->s, uid, 0444, (void *)bufnew(p->t, strlen(p->t))));
	if(s) closefile(d);
}

void
mkhdr(File *d, Hdr *h)
{
	mkpair(d, h->a, nil);
	mkpair(d, h->cd, "content-disposition");
	mkpair(d, h->ct, "content-type");
}

void
mkmulti(int n)
{
	File *multi, *d;
	Hdr hdr;
	char *s, buf[128];
	long l;

	for(; nmultipart<n; nmultipart++){
		incref(fs.tree->root);
		multi = walkfile(fs.tree->root, "request/multipart");
		snprint(buf, sizeof(buf), "%d", nmultipart);
		d = createfile(multi, buf, uid, 0555|DMDIR, (void *)0);
		parsemultipart(&hdr, nmultipart, &s, &l);
		closefile(createfile(d, "body", uid, 0444, (void *)bufnew(s, l)));
		mkhdr(d, &hdr);
		closefile(d);
		closefile(multi);
	}
}

HSPairs *
parseform(void)
{
	HSPairs *p;
	Buf *b;
	char *s;

	if(strcmp(cgihc->req.meth, "POST")!=0 || 
		(s=getval(reqhdr.ct, "type"))==nil || strcmp(s, Tformenc)!=0)
		return nil;

	alarm(HSTIMEOUT);
	while(getpost())
		;
	alarm(0);
	b = bufget(Xreqbody);
	p = hparsequery(cgihc, hstrdup(cgihc, b->s));
	return p;
}

void
inithttp(void)
{
	File *req, *auth;
	HSPriv *hp;
	HSPairs *p;

	parseheader(&reqhdr, (char *)cgihc->header, cgihc->hstop-cgihc->header);
	incref(fs.tree->root);
	req = walkfile(fs.tree->root, "request");
	mkhdr(req, &reqhdr);
	if(cgihc->req.search){
		bufset(Xreqquery, cgihc->req.search, strlen(cgihc->req.search));
		mkpair(req, hparsequery(cgihc, hstrdup(cgihc, cgihc->req.search)), "param");
	}

	if(cgihc->head.authuser)
		bufset(Xrequser, cgihc->head.authuser, strlen(cgihc->head.authuser));
	if(cgihc->head.authpass)
		bufset(Xreqpass, cgihc->head.authpass, strlen(cgihc->head.authpass));
	if(cgihc->head.authinfo){
		incref(req);
		auth = walkfile(req, "auth");
		mkpair(auth, cgihc->head.authinfo, nil);
	}

	hp = cgihc->private;
	bufset(Xreqrmtaddr, hp->remotesys, strlen(hp->remotesys));
	parsepath();
	bufset(Xreqmethod, cgihc->req.meth, strlen(cgihc->req.meth));
	if((p=parseform()) != nil)
		mkpair(req, p, "form");
	mkpair(req, cgihc->head.cookie, "cookie");
}

void
hwritebuf(Hio *h, Buf *b)
{
	long n, left;
	char *s, *p;

	if(b->s == nil)
		return;
	s = b->s;
	left = b->n;
	for(;;){
		if(left<=0 || (p=strchr(s, '\n'))==nil)
			break;
		n = (long)(p-s);
		hprint(h, "%.*s\r\n", (int)n, s);
		s +=n+1;
		left -= n+1;
	}
	/* not a newline at the end */
	if(left > 0) hprint(h, "%s\r\n", s);
}

void
outhttp(void)
{
	Buf *bd, *ct, *rp;
	Hio *h;

	h = &cgihc->hout;

	ct=bufget(Xresctype);
	if(ct->s == nil)
		error("no content-type");

	rp = bufget(Xresreply);
	if(rp->s == nil)
		error("no reply");

	bd = bufget(Xresbody);
	hprint(h, "%s ", hversion);
	hwritebuf(h, rp);
	hprint(h, "Server: Plan9\r\n");
	hprint(h, "Content-Length: %ld\r\n",bd->n);
	if(cgihc->head.closeit)
		hprint(h, "Connection: close\r\n");
	else if(!http11(cgihc))
		hprint(h, "Connection: Keep-Alive\r\n");
	hprint(h, "Content-type: ");
	hwritebuf(h, ct);

	hwritebuf(h, bufget(Xrescookie));
	hwritebuf(h, bufget(Xresnph));
	hprint(h, "\r\n");
	hwrite(h , bd->s, bd->n);
	hflush(h);
}

void
fsread(Req *r)
{
	int i;

	i = (uintptr)r->fid->file->aux;
	if(i < nbuffer){
		readbuf(r, buffer[i].s, buffer[i].n);
		respond(r, nil);
		return;
	}
	respond(r, "unknown read");
}

void
bufwrite(Req *r, Buf *b)
{
	void *v;
	vlong offset;
	long count;

	offset = r->ifcall.offset;
	count = r->ifcall.count;
	if(offset+count >= b->n){
		v = erealloc9p(b->s, offset+count+1);
		if(v == nil){
			respond(r, "no memory");
			return;
		}
		b->s = v;
		b->n = offset+count;
		b->s[b->n] = 0;
		r->fid->file->length = b->n;
	}
	memmove(b->s+offset, r->ifcall.data, count);
	r->ofcall.count = count;
}

static int
basicauth(void)
{
	char *uid, *pass;
	int n;
	AuthInfo *ai;

	uid = cgihc->head.authuser;
	pass = cgihc->head.authpass;

	n = 0;
	if (uid && pass) {
		ai = auth_userpasswd(uid, pass);
		if(ai && ai->cap)
			n = 1;

		auth_freeAI(ai);
	}
	return n;
}

static int
isauthattr(char *s)
{
	int m, l, h, c;

	static char *aih[] = {	/* must be sorted */
		"algorithm",
		"cnonce",
		"nc",
		"nonce",
		"opaque",
		"qop",
		"realm",
		"response",
		"uri",
		"username",
	};

	l = 0;
	h = (sizeof(aih)/sizeof(aih[0]));
	do {
		m = (l+h)/2;
		c = strcmp(s, aih[m]);
		if (c < 0) h = m-1;
		else if (c > 0) l = m+1;
		else return 1;
	} while (l <= h);

	return 0;
}

static char *
attrconcat(HSPairs *p)
{
	String *a;
	char *l;

	a = s_new();
	for (; p; p = p->next) {
		if (isauthattr(p->s)) {
			s_append(a, p->s);
			s_append(a, "=");
			s_append(a, p->t);
			if (p->next)
				s_append(a, ",");
		}
	}
	l = hstrdup(cgihc, s_to_c(a));
	s_free(a);
	return l;
}

static int
digestauth(void)
{
	char *f, buf[8*1024], *opaque = nil;
	HSPairs *aip;
	int fd, res = 0;

	aip = cgihc->head.authinfo;

	for (;aip != nil; aip = aip->next) {
		if (strcmp(aip->s, "opaque") == 0)
			opaque = aip->t;
	}

	if (opaque) {
		f = smprint("/mnt/digestauth/%s", opaque);
		fd = open(f, ORDWR);
		free(f);
		if (fd >= 0) {
			f = attrconcat(cgihc->head.authinfo);
			res =	write(fd, f, strlen(f)) == strlen(f) &&
				read(fd, buf, sizeof(buf)) > 0 &&
				strcmp(buf, "OK") == 0;
			close(fd);
		}
	}
	return res;
}

enum
{
	CMultipart,
	CAuthorize,
};

static Cmdtab cmdtab[] =
{
	CMultipart,	"multipart",	2,
	CAuthorize,	"authorize",	2,
};

void
fswrite(Req *r)
{
	Cmdbuf *cb;
	Cmdtab *ct;
	char *s;
	int i, n;

	i = (uintptr)r->fid->file->aux;
	if(i == Xctl){
		cb = parsecmd(r->ifcall.data, r->ifcall.count);
		ct = lookupcmd(cb, cmdtab, nelem(cmdtab));
		if(ct == nil){
			respondcmderror(r, cb, "unknown ctl command");
			return;
		}
		switch(ct->index){
		case CMultipart:
			if((n=atoi(cb->f[1])) < 0)
				break;
			if(strcmp(cgihc->req.meth, "POST") != 0){
				respondcmderror(r, cb, "no multiparts with GET");
				return;
			}
			if((s=getval(reqhdr.ct, "type")) == nil){
				respondcmderror(r, cb, "no content-type on request");
				return;
			}
			if(strcmp(s, Tmultipart) != 0){
				respondcmderror(r, cb, "content-type is no %s", Tmultipart);
				return;
			}
			mkmulti(n);
			respond(r, nil);
			return;
		case CAuthorize:
			if(cistrcmp(cb->f[1], "basic") == 0) {
				if(basicauth()) respond(r, nil);
				else respondcmderror(r, cb, "not authorized");
			}
			else if(cistrcmp(cb->f[1], "digest") == 0) {
				if (digestauth()) respond(r, nil);
				else respondcmderror(r, cb, "not authorized");
			}
			else
				respondcmderror(r, cb, "unknown auth type");
			return;
		}
		respondcmderror(r, cb, "bad cgi control message");
		return;
	}
	if(i == Xlog){
		logit(cgihc, "%s: %.*s", bufget(Xreqpath)->s, r->ifcall.count, r->ifcall.data);
		respond(r, nil);
		return;
	}
	if(i < nbuffer){
		bufwrite(r, bufget(i));
		respond(r, nil);
		return;
	}
	respond(r, "unknown write");
}

void
initfs(void)
{
	fs.tree = alloctree(uid, uid, DMDIR|0555, nil);
	mkdirtab(fs.tree->root, roottab);
}

void
main(int argc, char *argv[])
{
	Waitmsg *w;
	Hio *h;
	int i, pid;

	rfork(RFNOTEG);
	cgihc = init(argc, argv);
	h = &cgihc->hout;
	if(hparseheaders(cgihc, HSTIMEOUT) < 0)
		exits("failed");

	/* what do these do? */
	if(cgihc->head.expectother){
		hfail(cgihc, HExpectFail, nil);
		exits("failed");
	}
	if(cgihc->head.expectcont){
		hprint(h, "100 Continue\r\n");
		hprint(h, "\r\n");
		hflush(h);
	}

	nbuffer = NBUFS;
	buffer = emalloc9p(nbuffer*sizeof(Buf));

	initfs();
	inithttp();
	postmountsrv(&fs, nil, cgimtpt, MREPL);

	switch((pid=fork())){
	case -1:
		error("can't fork");

	case 0:
		for(i=0; i<20; i++)
			close(i);

		chdir(cgimtpt);
		execl("/bin/rc", "rc", "-c", bufget(Xreqpath)->s, 0);
		error("can't exec");
	}
	if((w = wait()) == nil || w->pid != pid)
		error("wait failed");
	if(strstr(w->msg, Eexec))
		error("exec failed");

	outhttp();
}
